#!/usr/bin/env ruby

#
# snarf_mail.rb
#
# Copies mail from an IMAP account
#
# Handy reference:
# http://ruby-doc.org/stdlib/libdoc/net/imap/rdoc/index.html
#

require 'date'
require 'net/imap'
require 'optparse'
require 'ostruct'

# Check that we were passed a date in %e-%b-%Y format.
def check_date(input)
  d = Date.parse(input)
  reserialized = d.strftime("%e-%b-%Y")
  reserialized.strip!
  if (reserialized != input) then
    raise "Invalid date '" + input + "': reserialized form was " + \
      "'" + reserialized + "', which does not match original."
  end
end

class MyOptions
  def self.parse(args)
    opts = OpenStruct.new
    opts.mailboxes = Array.new
    opts.since = nil
    opts.before = nil

    # Fill in $opts values
    parser = OptionParser.new do |myparser|
      myparser.banner = "Usage: #{ File.basename($0) } [opts]"
      myparser.separator("Specific options:")
      myparser.on("--since [TIME]", "-1",
              "Time to fetch email messages since.  (example: \
 1-Dec-2010.)") do |u|
        opts.since = u
        check_date(opts.since)
      end
      myparser.on("--before [TIME]", "-2",
              "Time to fetch email messages before.  (example: \
 1-Dec-2011.)") do |u|
        opts.before = u
        check_date(opts.before)
      end
      myparser.on("--username [USERNAME]", "-u",
              "Email account to fetch. (example: \
RareCactus@gmail.com or cmccabe@company.com)") do |u|
        opts.username = u
      end
      myparser.on("--list-folders", "-l",
              "List the IMAP folders that are present.") do |a|
        raise "can only specify one action" if (opts.action)
        opts.action = :list
      end
      myparser.on("--snarf", "-S",
              "Copy mail to the current directory.") do |a|
        raise "can only specify one action" if (opts.action)
        opts.action = :snarf
      end
      myparser.on("--box [MAILBOX]", "-b",
              "Act on a given mailbox. You may specify -b more than once for \
multiple mailboxes.") do |a|
        opts.mailboxes << a
      end
      myparser.on("--server [SERVER]", "-s",
              "Email server to use.  Example: imap.gmail.com") do |u|
        opts.server = u
      end
    end

    parser.parse!(args)
    raise "must specify an action" unless opts.action
    raise "must give a username" unless opts.username
    raise "must give a server" unless opts.server
    if ((opts.action == :snarf) and opts.mailboxes.empty?()) then
      raise "you must specify mailbox(es) with -b in order to snarf"
    end
    return opts
  end
end

# Get a password from STDIN without echoing it.
# This is kind of ugly, but it does work.
def get_password(prompt)
  shell_cmds = 'stty -echo && read password && echo ${password}'
  printf "#{prompt}"
  STDOUT.flush
  pass = ""
  pipe = IO.popen(shell_cmds, "r") do |pipe|
    pass = pipe.read
  end
  echo_status = $?.exitstatus
  system("stty sane")
  puts
  if (echo_status != 0) then
    raise "get_password: error executing: #{shell_cmds}"
  end
  return pass.chomp
end

def format_uid(uid)
  # We don't know how to deal with non-numeric UIDs. Best just to leave them
  # alone.
  return uid if (uid =~ /[^0123456789]/)

  # Pad numeric uids out to 6 digits
  return sprintf("%006d", uid)
end

def format_date(date)
  date.gsub!(' ', '_')
end

def get_sanitized_email_name(mailbox, arr)
  msn = mailbox.dup
  msn.gsub!(' ', '_')
  msn.gsub!('/', '.')
  return "#{msn}_#{format_date(arr["INTERNALDATE"])}_#{format_uid(arr["UID"])}"
end

def write_email_to_disk(mailbox, data)
  arr = data[0].attr
  filename = get_sanitized_email_name(mailbox, arr)
  fp = File.open(filename, 'w')
  fp.write(arr["RFC822.HEADER"])
  fp.write(arr["RFC822.TEXT"])
  fp.close
end

def snarf_mailbox(imap, mailbox)
  full_count = 0
  first_time = true

  searchterms = [ "NOT", "DELETED" ]
  if $opts.since != nil then
    searchterms << "SINCE" << $opts.since
  end
  if $opts.before != nil then
    searchterms << "BEFORE" << $opts.before
  end
  print "using IMAP search terms: " + searchterms.join(" ")
  prequel = "fetched: "

  while true
    count = 0
    msg_seqnos = Array.new

    imap.select(mailbox)
    imap.search(searchterms).each do |message_id|
      if (first_time == true) then
        # Print a dot immediately after making first contact with the server.
        # It is reassuring to the user.
        printf(".")
        STDOUT.flush()
        first_time = false
      end
      data = imap.fetch(message_id, 
                [ "INTERNALDATE", "UID", "RFC822.HEADER", "RFC822.TEXT" ])
      write_email_to_disk(mailbox, data)
      count = count + 1
      full_count = full_count + 1
      msg_seqnos << data[0].seqno.to_i
      #break if (count > 20)
    end
    if (count == 0) then
      puts "#{prequel} #{full_count} messages from #{mailbox}"
      return
    end

    # Print out a dot to signify progress
    printf(".")
    STDOUT.flush()

    if $opts.delete != "none" then
      # Delete messages
      imap.store(msg_seqnos, "+FLAGS", [:Deleted])
      imap.expunge
    end
  end
end

# MAIN
begin
  $opts = MyOptions.parse(ARGV)
rescue Exception => msg
  $stderr.print("#{msg}.\nType --help to see usage information.\n")
  exit 1
end

password = get_password("Please enter the password for #{$opts.username}:")
imap = Net::IMAP.new($opts.server, 993, true)
imap.login($opts.username, password)
case ($opts.action)
when :list
  imap.list("", "*").each do |mbl|
    puts "#{mbl.name}"
  end
when :snarf
  $opts.mailboxes.each do |mailbox|
    snarf_mailbox(imap, mailbox)
  end
else
  raise "unknown action #{$opts.action}"
end
imap.logout()
imap.disconnect()
exit 0
